"""Test cli sql module."""
from click.testing import CliRunner
import json

from ladybug.datacollection import HourlyContinuousCollection
from ladybug_comfort.parameter.adaptive import AdaptiveParameter
from ladybug_comfort.cli.sql import pmv_by_room, adaptive_by_room


def test_pmv_by_room():
    runner = CliRunner()
    input_sql = './tests/sql/eplusout.sql'

    result = runner.invoke(pmv_by_room, [input_sql])
    assert result.exit_code == 0
    data_dicts = json.loads(result.output)
    pmv_data = [HourlyContinuousCollection.from_dict(dat) for dat in data_dicts]
    assert len(pmv_data) == 2
    assert len(pmv_data[0]) == 8760


def test_adaptive_by_room():
    runner = CliRunner()
    input_sql = './tests/sql/eplusout.sql'
    input_epw = './tests/epw/chicago.epw'

    result = runner.invoke(adaptive_by_room, [input_sql, input_epw, '-v', '0.65'])
    assert result.exit_code == 0
    data_dicts = json.loads(result.output)
    ad_data = [HourlyContinuousCollection.from_dict(dat) for dat in data_dicts]
    assert len(ad_data) == 2
    assert len(ad_data[0]) == 8760


def test_adaptive_by_room_custom():
    runner = CliRunner()
    input_sql = './tests/sql/eplusout.sql'
    input_epw = './tests/epw/chicago.epw'

    air_speed = [0.65] * 8760
    air_speed = json.dumps(air_speed)
    comf_par = AdaptiveParameter(False)
    comf_par.set_neutral_offset_from_comfort_class(1)

    result = runner.invoke(
        adaptive_by_room, [input_sql, input_epw, '-v', air_speed, '-cp', str(comf_par)])
    assert result.exit_code == 0
    data_dicts = json.loads(result.output)
    ad_data = [HourlyContinuousCollection.from_dict(dat) for dat in data_dicts]
    assert len(ad_data) == 2
    assert len(ad_data[0]) == 8760



"""Run EPW weather data through comfort models."""
import click
import sys
import logging
import json

from ladybug.epw import EPW
from ladybug_comfort.collection.utci import UTCI
from ladybug_comfort.collection.pmv import PMV

_logger = logging.getLogger(__name__)


@click.group(help='Commands for running EPW weather data through comfort models.')
def epw():
    pass


@epw.command('utci')
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--include-wind/--exclude-wind', ' /-xw', help='Flag to note whether '
              'to include the EPW wind speed in the calculation.',
              default=True, show_default=True)
@click.option('--include-sun/--exclude-sun', ' /-xs', help='Flag to note whether '
              'to include the mean radiant temperature (MRT) delta from both shortwave '
              'solar falling directly on people and long wave radiant exchange with '
              'the sky.', default=True, show_default=True)
@click.option('--csv/--json', help='Flag to note whether output data collection should '
              'be in JSON format.', default=True, show_default=True)
@click.option('--output-file', '-f', help='Optional file to output the CSV or JSON '
              'string of the data. By default, it will be printed to stdout',
              type=click.File('w'), default='-', show_default=True)
def utci(epw_file, include_wind, include_sun, csv, output_file):
    """Get a data collection for UTCI from an EPW weather file.

    \b
    Args:
        epw_file: Path to an .epw file.
    """
    try:
        epw_obj = EPW(epw_file)
        utci_obj = UTCI.from_epw(epw_obj, include_wind, include_sun)
        utci_data = utci_obj.universal_thermal_climate_index
        if csv:
            output_file.write('\n'.join([str(v) for v in utci_data.values]))
        else:
            output_file.write(json.dumps(utci_data.to_dict()))
    except Exception as e:
        _logger.exception('Failed to get UTCI from EPW file.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@epw.command('set')
@click.argument('epw-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--include-wind/--exclude-wind', ' /-xw', help='Flag to note whether '
              'to include the EPW wind speed in the calculation.',
              default=True, show_default=True)
@click.option('--include-sun/--exclude-sun', ' /-xs', help='Flag to note whether '
              'to include the mean radiant temperature (MRT) delta from both shortwave '
              'solar falling directly on people and long wave radiant exchange with '
              'the sky.', default=True, show_default=True)
@click.option('--met-rate', '-m', help='A number for metabolic rate in met.',
              type=float, default=2.4, show_default=True)
@click.option('--clo-value', '-c', help='A number for clothing level in clo.',
              type=float, default=0.7, show_default=True)
@click.option('--csv/--json', help='Flag to note whether output data collection should '
              'be in JSON format.', default=True, show_default=True)
@click.option('--output-file', '-f', help='Optional file to output the JSON string of '
              'the data collection. By default, it will be printed to stdout',
              type=click.File('w'), default='-', show_default=True)
def set_(epw_file, include_wind, include_sun, met_rate, clo_value, csv, output_file):
    """Get a data collection for Standard Effective Temperature from an EPW weather file.

    \b
    Args:
        epw_file: Path to an .epw file.
    """
    try:
        epw_obj = EPW(epw_file)
        pmv_obj = PMV.from_epw(epw_obj, include_wind, include_sun, met_rate, clo_value)
        set_data = pmv_obj.standard_effective_temperature
        if csv:
            output_file.write('\n'.join([str(v) for v in set_data.values]))
        else:
            output_file.write(json.dumps(set_data.to_dict()))
    except Exception as e:
        _logger.exception('Failed to get SET from EPW file.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)



"""Test cli epw module."""
from click.testing import CliRunner
import json

from ladybug.datacollection import HourlyContinuousCollection
from ladybug_comfort.cli.epw import utci, set_


def test_utci():
    runner = CliRunner()
    input_epw = './tests/epw/chicago.epw'

    result = runner.invoke(
        utci, [input_epw, '--exclude-wind', '--exclude-sun', '--csv'])
    assert result.exit_code == 0

    result = runner.invoke(
        utci, [input_epw, '--exclude-wind', '--exclude-sun', '--json'])
    assert result.exit_code == 0
    data_dict = json.loads(result.output)
    utci_data = HourlyContinuousCollection.from_dict(data_dict)
    assert len(utci_data) == 8760


def test_set():
    runner = CliRunner()
    input_epw = './tests/epw/chicago.epw'

    result = runner.invoke(
        set_, [input_epw, '--exclude-wind', '--exclude-sun', '--json'])
    assert result.exit_code == 0
    data_dict = json.loads(result.output)
    set_data = HourlyContinuousCollection.from_dict(data_dict)
    assert len(set_data) == 8760



# coding=utf-8
"""Methods for resolving MRT from Radiance output files."""
from __future__ import division

from ladybug.sunpath import Sunpath
from ladybug.datatype.energyflux import Irradiance
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.header import Header
from ladybug.datacollection import HourlyContinuousCollection

from ..solarcal import sharp_from_solar_and_body_azimuth
from ..collection.solarcal import _HorizontalSolarCalMap, _HorizontalRefSolarCalMap
from ..parameter.solarcal import SolarCalParameter


def shortwave_mrt_map(location, longwave_data, sun_up_hours, total_ill, direct_ill,
                      ref_ill=None, solarcal_par=None):
    """Get MRT data collections adjusted for shortwave using Radiance .ill files.

    Args:
        location: A ladybug Location object to dictate the solar positions used
            in the calculation.
        longwave_data: An array of data collections for each point within a thermal
            map. All collections must be aligned with one another. The analysis
            period on the collections does not have to be annual.
        sun_up_hours: File path to a sun-up-hours.txt file output by Radiance.
        total_ill: Path to an .ill file output by Radiance containing total
            irradiance for each longwave_data collection.
        direct_ill: Path to an .ill file output by Radiance containing direct
            irradiance for each longwave_data collection.
        ref_ill: Path to an .ill file output by Radiance containing total ground-
            reflected irradiance for each longwave_data collection. If None, a
            default ground reflectance of 0.25 will be assumed.
        solarcal_par: Optional SolarCalParameter object to account for
            properties of the human geometry.
    """
    # determine the analysis period and open the sun_up_hours file
    a_per = longwave_data[0].header.analysis_period
    is_annual, t_step, lp_yr = a_per.is_annual, a_per.timestep, a_per.is_leap_year
    with open(sun_up_hours) as soh_f:
        sun_indices = [int(float(h) * t_step) for h in soh_f]

    # parse each of the .ill files into annual irradiance data collections
    total = _ill_file_to_data(total_ill, sun_indices, t_step, lp_yr)
    direct = _ill_file_to_data(direct_ill, sun_indices, t_step, lp_yr)
    ref = _ill_file_to_data(ref_ill, sun_indices, t_step, lp_yr) \
        if ref_ill is not None else None

    # if the analysis is not annual, apply analysis periods
    if not is_annual:
        total = [data.filter_by_analysis_period(a_per) for data in total]
        direct = [data.filter_by_analysis_period(a_per) for data in direct]
        if ref is not None:
            ref = [data.filter_by_analysis_period(a_per) for data in ref]

    # convert total irradiance into indirect irradiance
    indirect = [t_rad - d_rad for t_rad, d_rad in zip(total, direct)]

    # compute solar altitudes and sharps
    body_par = SolarCalParameter() if solarcal_par is None else solarcal_par
    sp = Sunpath.from_location(location)
    _altitudes = []
    if body_par.body_azimuth is None:
        _sharps = [body_par.sharp] * len(a_per)
        for t_date in a_per.datetimes:
            sun = sp.calculate_sun_from_date_time(t_date)
            _altitudes.append(sun.altitude)
    else:
        _sharps = []
        for t_date in a_per.datetimes:
            sun = sp.calculate_sun_from_date_time(t_date)
            sharp = sharp_from_solar_and_body_azimuth(sun.azimuth, body_par.body_azimuth)
            _sharps.append(sharp)
            _altitudes.append(sun.altitude)

    # pass all data through the solarcal collections and return MRT data collections
    mrt_data = []
    if ref is not None:  # fully-detailed SolarCal with ground reflectance
        for l_mrt, d_rad, i_rad, r_rad in zip(longwave_data, direct, indirect, ref):
            scl_obj = _HorizontalRefSolarCalMap(
                _altitudes, _sharps, d_rad, i_rad, r_rad, l_mrt, None, body_par)
            mrt_data.append(scl_obj.mean_radiant_temperature)
    else:  # simpler SolarCal assuming default ground reflectance
        for l_mrt, d_rad, i_rad in zip(longwave_data, direct, indirect):
            scl_obj = _HorizontalSolarCalMap(
                _altitudes, _sharps, d_rad, i_rad, l_mrt, None, None, body_par)
            mrt_data.append(scl_obj.mean_radiant_temperature)
    return mrt_data


def _ill_file_to_data(ill_file, sun_indices, timestep=1, leap_yr=False):
    """Convert a list of sun-up irradiance from an .ill file into annual irradiance data.

    Args:
        ill_values: A list of raw irradiance values from an .ill file.
        sun_indices: A list of integers for where in the total_count sun-up hours occur.
        timestep: The timestep to make the data collection.
        leap_yr: Boolean to note if data is for a leap year.
    
    Return:
        A list of annual HourlyContinuousCollection with irradiance data.
    """
    irr_data = []
    with open(ill_file) as results:
        for pt_res in results:
            ill_values = [float(v) for v in pt_res.split()]
            pt_irr_data = _ill_values_to_data(ill_values, sun_indices, timestep, leap_yr)
            irr_data.append(pt_irr_data)
    return irr_data


def _ill_values_to_data(ill_values, sun_indices, timestep=1, leap_yr=False):
    """Convert a list of sun-up irradiance from an .ill file into annual irradiance data.

    Args:
        ill_values: A list of raw irradiance values from an .ill file.
        sun_indices: A list of integers for where in the total_count sun-up hours occur.
        timestep: The timestep to make the data collection.
        leap_yr: Boolean to note if data is for a leap year.
    
    Return:
        An annual HourlyContinuousCollection of irradiance data.
    """
    values = [0] * (8760 * timestep) if not leap_yr else [0] * (8784 * timestep)
    for i, irr in zip(sun_indices, ill_values):
        values[i] = irr
    a_period = AnalysisPeriod(timestep=timestep, is_leap_year=leap_yr)
    header = Header(Irradiance(), 'W/m2', a_period)
    return HourlyContinuousCollection(header, values)



# coding utf-8
from ladybug_comfort.map.mrt import shortwave_mrt_map
from ladybug_comfort.map._enclosure import _parse_enclosure_info

from ladybug.datacollection import HourlyContinuousCollection
from ladybug.sql import SQLiteResult
from ladybug.epw import EPW


# global files object used by all of the tests
sun_up_path = './tests/map/results/total/sun-up-hours.txt'
total_ill_path = './tests/map/results/total/TestRoom_1.ill'
direct_ill_path = './tests/map/results/direct/TestRoom_1.ill'
ref_ill_path = './tests/map/results/total/TestRoom_1_ref.ill'
enclosure_path = './tests/map/TestRoom_1_enclosure.json'
sql_path = './tests/sql/eplusout.sql'
sql = SQLiteResult(sql_path)
epw_path = './tests/epw/boston.epw'
epw = EPW(epw_path)


def test_shortwave_mrt_map():
    """Test the shortwave_mrt_map method."""
    location = epw.location
    l_mrt_data = sql.data_collections_by_output_name('Zone Mean Radiant Temperature')
    l_mrt_data = [l_mrt_data[0]] * 4

    mrt_map_data = shortwave_mrt_map(
        location, l_mrt_data, sun_up_path, total_ill_path, direct_ill_path, ref_ill_path)

    assert len(mrt_map_data) == 4
    for mrt_dat in mrt_map_data:
        assert isinstance(mrt_dat, HourlyContinuousCollection)
        assert len(mrt_dat) == 8760


def test_parse_enclosure_info():
    """Test the _parse_enclosure_info method."""
    pt_air_temps, pt_rad_temps, pt_humids, pt_speeds, a_period = _parse_enclosure_info(
        enclosure_path, sql_path, epw, include_humidity=True)

    assert len(pt_air_temps) == 4
    for air_dat in pt_air_temps:
        assert isinstance(air_dat, HourlyContinuousCollection)
        assert len(air_dat) == 8760
    assert len(pt_rad_temps) == 4
    for mrt_dat in pt_rad_temps:
        assert isinstance(mrt_dat, HourlyContinuousCollection)
        assert len(mrt_dat) == 8760
    assert len(pt_humids) == 4
    for hum_dat in pt_humids:
        assert isinstance(hum_dat, HourlyContinuousCollection)
        assert len(hum_dat) == 8760
    assert len(pt_speeds) == 4



"""Test cli sql module."""
from click.testing import CliRunner
import json
import os

from ladybug.futil import nukedir
from ladybug.header import Header
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.datatype.temperature import OperativeTemperature, \
    StandardEffectiveTemperature, UniversalThermalClimateIndex
from ladybug.datatype.thermalcondition import PredictedMeanVote, \
    ThermalCondition, ThermalConditionElevenPoint
from ladybug.datatype.temperaturedelta import OperativeTemperatureDelta

from ladybug_comfort.cli.map import pmv, adaptive, utci, map_result_info, tcp

# global files object used by all of the tests
sql_path = './tests/sql/eplusout.sql'
enclosure_path = './tests/map/TestRoom_1_enclosure.json'
epw_path = './tests/epw/boston.epw'
sun_up_path = './tests/map/results/total/sun-up-hours.txt'
total_ill_path = './tests/map/results/total/TestRoom_1.ill'
direct_ill_path = './tests/map/results/direct/TestRoom_1.ill'
ref_ill_path = './tests/map/results/total/TestRoom_1_ref.ill'
enclosure_path = './tests/map/TestRoom_1_enclosure.json'


def test_pmv_map():
    runner = CliRunner()
    res_folder = './tests/map/pmv_map_results'
    run_period = AnalysisPeriod(1, 2, 0, 1, 2, 23)

    base_cmd = [sql_path, enclosure_path, epw_path]
    base_cmd.extend(['-tr', total_ill_path, '-dr', direct_ill_path, '-rr', ref_ill_path])
    base_cmd.extend(['-sh', sun_up_path])
    base_cmd.extend(['-rp', str(run_period)])
    base_cmd.extend(['--folder', res_folder])

    result = runner.invoke(pmv, base_cmd)

    assert result.exit_code == 0
    out_files = json.loads(result.output)
    assert os.path.isfile(out_files['temperature'])
    assert os.path.isfile(out_files['condition'])
    assert os.path.isfile(out_files['condition_intensity'])

    nukedir(res_folder, True)


def test_adaptive_map():
    runner = CliRunner()
    res_folder = './tests/map/adaptive_map_results'

    base_cmd = [sql_path, enclosure_path, epw_path]
    base_cmd.extend(['-tr', total_ill_path, '-dr', direct_ill_path, '-rr', ref_ill_path])
    base_cmd.extend(['-sh', sun_up_path])
    base_cmd.extend(['--folder', res_folder])

    result = runner.invoke(adaptive, base_cmd)

    assert result.exit_code == 0
    out_files = json.loads(result.output)
    assert os.path.isfile(out_files['temperature'])
    assert os.path.isfile(out_files['condition'])
    assert os.path.isfile(out_files['condition_intensity'])

    nukedir(res_folder, True)


def test_utci_map():
    runner = CliRunner()
    res_folder = './tests/map/utci_map_results'

    base_cmd = [sql_path, enclosure_path, epw_path]
    base_cmd.extend(['-tr', total_ill_path, '-dr', direct_ill_path, '-rr', ref_ill_path])
    base_cmd.extend(['-sh', sun_up_path])
    base_cmd.extend(['--folder', res_folder])

    result = runner.invoke(utci, base_cmd)

    assert result.exit_code == 0
    out_files = json.loads(result.output)
    assert os.path.isfile(out_files['temperature'])
    assert os.path.isfile(out_files['condition'])
    assert os.path.isfile(out_files['condition_intensity'])

    nukedir(res_folder, True)


def test_map_result_info():
    runner = CliRunner()
    a_per = AnalysisPeriod()
    a_per_sub = AnalysisPeriod(6, 21, 0, 9, 21, 23)

    cmd = ['pmv', '--run-period', '', '--qualifier', 'write-op-map']
    result = runner.invoke(map_result_info, cmd)
    assert result.exit_code == 0
    out_files = json.loads(result.output)
    assert out_files['temperature'] == \
        Header(OperativeTemperature(), 'C', a_per).to_dict()
    assert out_files['condition'] == \
        Header(ThermalCondition(), 'condition', a_per).to_dict()
    assert out_files['condition_intensity'] == \
        Header(PredictedMeanVote(), 'PMV', a_per).to_dict()

    cmd = ['pmv', '--run-period', str(a_per_sub), '--qualifier', 'write-set-map']
    result = runner.invoke(map_result_info, cmd)
    assert result.exit_code == 0
    out_files = json.loads(result.output)
    assert out_files['temperature'] == \
        Header(StandardEffectiveTemperature(), 'C', a_per_sub).to_dict()

    cmd = ['pmv']
    result = runner.invoke(map_result_info, cmd)
    assert result.exit_code == 0

    cmd = ['adaptive', '--run-period', '', '--qualifier', '']
    result = runner.invoke(map_result_info, cmd)
    assert result.exit_code == 0
    out_files = json.loads(result.output)
    assert out_files['temperature'] == \
        Header(OperativeTemperature(), 'C', a_per).to_dict()
    assert out_files['condition_intensity'] == \
        Header(OperativeTemperatureDelta(), 'dC', a_per).to_dict()

    cmd = ['utci']
    result = runner.invoke(map_result_info, cmd)
    assert result.exit_code == 0
    out_files = json.loads(result.output)
    assert out_files['temperature'] == \
        Header(UniversalThermalClimateIndex(), 'C', a_per).to_dict()
    assert out_files['condition_intensity'] == \
        Header(ThermalConditionElevenPoint(), 'condition', a_per).to_dict()


def test_tcp():
    runner = CliRunner()
    condition_path = './tests/map/map_results/condition.csv'
    occ_sch_path = './tests/map/occ_schedules.json'
    res_folder = './tests/map/metrics'

    base_cmd = [condition_path, enclosure_path, '--occ-schedule-json', occ_sch_path]
    base_cmd.extend(['--folder', res_folder])
    result = runner.invoke(tcp, base_cmd)

    assert result.exit_code == 0
    out_files = json.loads(result.output)
    for fp in out_files:
        assert os.path.isfile(fp)
    nukedir(res_folder, True)

    cmds = [condition_path, enclosure_path, '--folder', res_folder]
    result = runner.invoke(tcp, cmds)

    assert result.exit_code == 0
    out_files = json.loads(result.output)
    for fp in out_files:
        assert os.path.isfile(fp)
    nukedir(res_folder, True)
